@@ -7,7 +7,7 @@ import shutil
 from subprocess import PIPE
 
 from common import RunnerCore, create_file, read_file, test_file
-from decorators import parameterized
+from decorators import also_with_wasm64, also_without_bigint, parameterized
 
 from tools.shared import EMCC
 from tools.utils import delete_file
@@ -286,8 +286,27 @@ addToLibrary({
     err = self.expect_fail([EMCC, test_file('hello_world.c'), '--js-library', 'lib.js'])
     self.assertContained("lib.js: Decorator (jslibfunc__async} has wrong type. Expected 'boolean' not 'string'", err)
 
+  @also_with_wasm64
+  @also_without_bigint
   def test_jslib_i53abi(self):
     create_file('lib.js', r'''
+mergeInto(LibraryManager.library, {
+  jslibfunc__i53abi: true,
+  jslibfunc__sig: 'j',
+  jslibfunc: (x) => { return 42 },
+});
+''')
+    create_file('test.c', r'''
+#include <stdio.h>
+int64_t jslibfunc();
+int main() {
+  printf("main: %lld\n", jslibfunc());
+}
+''')
+    self.do_runf('test.c', 'main: 42\n', cflags=['--js-library', 'lib.js'])
+
+  def test_jslib_i53abi_errors(self):
+    create_file('lib.js', r'''
 mergeInto(LibraryManager.library, {
   jslibfunc__i53abi: true,
   jslibfunc: (x) => { return 42 },
@@ -476,6 +495,8 @@ extraLibraryFuncs.push('jsfunc');
     err = self.expect_fail([EMCC, test_file('hello_world.c'), '--js-library', 'foo.js'])
     self.assertContained('foo.js:5: file not found: inc.js', err)
 
+  @also_with_wasm64
+  @also_without_bigint
   @parameterized({
     '': ([],),
     'closure': (['--closure=1'],),
@@ -484,8 +505,17 @@ extraLibraryFuncs.push('jsfunc');
     create_file('foo.js', '''
       addToLibrary({
         foo: () => 42,
+        foo__sig: 'i',
+
         foo_alias: 'foo',
 
+        foo_alias_i64: 'foo',
+        foo_alias_i64__sig: 'j',
+        foo_alias_i64__i53abi: true,
+
+        foo_alias_ptr: 'foo',
+        foo_alias_ptr__sig: 'p',
+
         // Normal JS function that calls a native function
         call_native__deps: ['native_func'],
         call_native: () => _native_func(),
@@ -505,6 +535,8 @@ extraLibraryFuncs.push('jsfunc');
       #include <emscripten.h>
       int foo();
       int foo_alias();
+      void* foo_alias_ptr();
+      int64_t foo_alias_i64();
       int call_native();
       int call_native_alias();
 
@@ -519,6 +551,8 @@ extraLibraryFuncs.push('jsfunc');
       int main() {
         printf("foo: %d\n", foo());
         printf("foo_alias: %d\n", foo_alias());
+        printf("foo_alias_i64: %lld\n", foo_alias_i64());
+        printf("foo_alias_ptr: %p\n", foo_alias_ptr());
         printf("call_native: %d\n", call_native());
         printf("call_native_alias: %d\n", call_native_alias());
         return 0;
@@ -527,6 +561,8 @@ extraLibraryFuncs.push('jsfunc');
     expected = '''\
 foo: 42
 foo_alias: 42
+foo_alias_i64: 42
+foo_alias_ptr: 0x2a
 call_native: 43
 call_native_alias: 44
 '''
@@ -671,3 +707,11 @@ console.error('JSLIB: none of the above');
     # When WebGL is explicitly linked to in strict mode, the linking order on command line should enable overriding.
     self.cflags += ['-sAUTO_JS_LIBRARIES=0', '-sMAX_WEBGL_VERSION=2', '-lwebgl.js', '--js-library', test_file('test_jslib_override_system_symbol.js')]
     self.do_run_in_out_file_test('test_jslib_override_system_symbol.c')
+
+  def test_jslib_version_check(self):
+    create_file('libfoo.js', '''
+      #if parseInt(EMSCRIPTEN_VERSION.split('.')[0]) > 3
+      #error "library does not support emscripten > 3.0.0"
+      #endif
+    ''')
+    self.assert_fail([EMCC, '--js-library=libfoo.js'], 'error: libfoo.js:3: #error "library does not support emscripten > 3.0.0"')
