@@ -14,9 +14,12 @@ import os
 import unittest
 
 import test_posixtest_browser
+from browser_common import browser_should_skip_feature
 from common import RunnerCore, path_from_root
 from decorators import node_pthreads
 
+from tools.feature_matrix import Feature
+
 testsuite_root = path_from_root('test/third_party/posixtestsuite')
 
 
@@ -161,6 +164,9 @@ def make_test(name, testfile, browser):
   def f(self):
     if name in disabled:
       self.skipTest(disabled[name])
+    if browser and browser_should_skip_feature('EMTEST_LACKS_SHARED_ARRAY_BUFFER', Feature.THREADS):
+      self.skipTest('This test requires a browser with SharedArrayBuffer support')
+
     args = ['-I' + os.path.join(testsuite_root, 'include'),
             '-Werror',
             '-Wno-format-security',
