@@ -8,6 +8,7 @@ import os
 import re
 import shutil
 import stat
+import sys
 import time
 from pathlib import Path
 from subprocess import PIPE, STDOUT
@@ -18,11 +19,12 @@ from common import (
   create_file,
   ensure_dir,
   env_modify,
+  exe_suffix,
   make_executable,
   path_from_root,
   test_file,
 )
-from decorators import crossplatform, parameterized, with_env_modify
+from decorators import crossplatform, no_windows, parameterized, with_env_modify
 
 from tools import cache, ports, response_file, shared, utils
 from tools.config import EM_CONFIG
@@ -30,7 +32,6 @@ from tools.shared import EMCC, config
 from tools.utils import delete_dir, delete_file
 
 SANITY_FILE = cache.get_path('sanity.txt')
-commands = [[EMCC], [shared.bat_suffix(path_from_root('test/runner')), 'blahblah']]
 expected_llvm_version = str(shared.EXPECTED_LLVM_VERSION) + '.0.0'
 
 
@@ -99,6 +100,22 @@ def make_fake_clang(filename, version, targets='wasm32 - WebAssembly 32-bit'):
   make_fake_tool(filename + '++', version, output)
 
 
+# Return a new PATH that has no directories that would contain the given tool.
+def path_without_tool(env_path, tool_bin):
+  tool_bin = exe_suffix(tool_bin)
+  python_path = os.path.normpath(os.path.dirname(sys.executable))
+
+  def ignore_path(p):
+    # We cannot ignore a path element that contains the python executable itself, otherwise
+    # the bootstrap script will fail
+    if os.path.isfile(os.path.join(p, tool_bin)) and os.path.normpath(p) != python_path:
+      return True
+    return False
+
+  old_path = env_path.split(os.pathsep)
+  return os.pathsep.join([d for d in old_path if not ignore_path(d)])
+
+
 SANITY_MESSAGE = 'Emscripten: Running sanity checks'
 
 # arguments to build a minimal hello world program, without even libc
@@ -152,33 +169,34 @@ class sanity(RunnerCore):
 
     return self.run_process(command, stdout=PIPE, stderr=STDOUT, check=False, env=env).stdout
 
-  def check_working(self, command, expected=None):
+  def check_working(self, command, expected=None, env=None):
     if type(command) is not list:
       command = [command]
     if command == [EMCC]:
       command = [EMCC, '--version']
     if expected is None:
-      if command[0] == EMCC or (len(command) >= 2 and command[1] == EMCC):
-        expected = 'emcc (Emscripten gcc/clang-like replacement + linker emulating GNU ld)'
-      else:
-        expected = 'could not find the following tests: blahblah'
+      expected = 'emcc (Emscripten gcc/clang-like replacement + linker emulating GNU ld)'
 
-    output = self.do(command)
+    output = self.do(command, env=env)
     self.assertContained(expected, output)
     return output
 
   # this should be the very first thing that runs. if this fails, everything else is irrelevant!
   @crossplatform
   def test_aaa_normal(self):
-    for command in commands:
-      # Your existing EM_CONFIG should work!
-      restore_and_set_up()
-      self.check_working(command)
+    # Your existing EM_CONFIG should work!
+    restore_and_set_up()
+    self.check_working(EMCC)
 
   @with_env_modify({'EM_CONFIG': None})
   def test_firstrun(self):
+    # Remove from PATH every directory that contains clang.exe so config setup
+    # cannot accidentally succeed by virtue of locating tools in PATH.
+    env = os.environ.copy()
+    env['PATH'] = path_without_tool(env['PATH'], 'clang')
+
     default_config = path_from_root('.emscripten')
-    output = self.do([EMCC, '-v'])
+    output = self.do([EMCC, '-v'], env=env)
     self.assertContained('emcc: warning: config file not found: %s.  You can create one by hand or run `emcc --generate-config`' % default_config, output)
 
     temp_bin = os.path.abspath('bin')
@@ -223,20 +241,20 @@ class sanity(RunnerCore):
     # output = self.run_process([EMCC, test_file('hello_world.c')], stdout=PIPE, stderr=PIPE).output
     # self.assertContained('hello, world!', self.run_js('a.out.js'), output)
 
-    for command in commands:
-      # Second run, with bad EM_CONFIG
-      for settings in ('blah', 'LLVM_ROOT="blarg"; JS_ENGINES=[]; NODE_JS=[]; SPIDERMONKEY_ENGINE=[]'):
-        try:
-          utils.write_file(default_config, settings)
-          output = self.do(command)
+    # Second run, with bad EM_CONFIG
+    for settings in ('blah', 'LLVM_ROOT="blarg"; JS_ENGINES=[]; NODE_JS=[]; SPIDERMONKEY_ENGINE=[]'):
+      try:
+        utils.write_file(default_config, settings)
+        output = self.do(EMCC)
 
-          if 'blah' in settings:
-            self.assertContained('error: error in evaluating config file (%s)' % default_config, output)
-          elif 'runner' not in ' '.join(command):
-            self.assertContained('error: NODE_JS is set to empty value', output) # sanity check should fail
-        finally:
-          delete_file(default_config)
+        if 'blah' in settings:
+          self.assertContained('error: error in evaluating config file (%s)' % default_config, output)
+        else:
+          self.assertContained('error: NODE_JS is set to empty value', output) # sanity check should fail
+      finally:
+        delete_file(default_config)
 
+  @no_windows('Test relies on Unix-specific make_fake_tool')
   def test_llvm(self):
     LLVM_WARNING = 'LLVM version for clang executable'
 
@@ -276,6 +294,7 @@ class sanity(RunnerCore):
           output = self.check_working(EMCC)
           self.assertNotContained(LLVM_WARNING, output)
 
+  @no_windows('Test relies on Unix-specific shell script')
   def test_node(self):
     NODE_WARNING = 'node version appears too old'
     NODE_WARNING_2 = 'cannot check node version'
@@ -373,10 +392,9 @@ fi
     ''')
 
     wipe()
+    expected = 'error: inline EM_CONFIG data no longer supported.  Please use a config file.'
     with env_modify({'EM_CONFIG': get_basic_config()}):
-      out = self.expect_fail([EMCC, 'main.cpp', '-Wno-deprecated', '-o', 'a.out.js'])
-
-    self.assertContained('error: inline EM_CONFIG data no longer supported.  Please use a config file.', out)
+      self.assert_fail([EMCC, 'main.cpp', '-Wno-deprecated', '-o', 'a.out.js'], expected)
 
   def clear_cache(self):
     self.run_process([EMCC, '--clear-cache'])
@@ -423,6 +441,7 @@ fi
     self.assertIn(SANITY_MESSAGE, output)
     self.assertCacheEmpty()
 
+  @no_windows('Test relies on Unix-specific make_fake_tool')
   def test_cache_clearing_auto(self):
     # Changing LLVM_ROOT, even without altering .emscripten, clears the cache
     restore_and_set_up()
@@ -585,6 +604,7 @@ fi
       first_use()
       second_use()
 
+  @no_windows('Test relies on Unix-specific shell script')
   def test_js_engine_path(self):
     # Test that running JS commands works for node, d8, and jsc and is not path dependent
     restore_and_set_up()
@@ -640,6 +660,7 @@ fi
       build()
       test()
 
+  @no_windows('Test relies on Unix-specific make_fake_tool')
   def test_vanilla(self):
     restore_and_set_up()
     self.clear_cache()
@@ -664,6 +685,7 @@ fi
     test_with_fake('got js backend! JavaScript (asm.js, emscripten) backend', 'LLVM has not been built with the WebAssembly backend')
     delete_dir(shared.CANONICAL_TEMP_DIR)
 
+  @no_windows('Test relies on Unix-specific make_fake_tool')
   def test_llvm_add_version(self):
     restore_and_set_up()
 
@@ -676,8 +698,7 @@ fi
       make_fake_tool(self.in_dir('fake', f'llvm-nm-{version}'), expected_llvm_version)
 
     make_fake('9.9')
-    out = self.expect_fail([EMCC, '-v'])
-    self.assertContained('No such file or directory', out)
+    self.assert_fail([EMCC, '-v'], 'No such file or directory')
     with env_modify({'EM_LLVM_ADD_VERSION': '9.9', 'EM_CLANG_ADD_VERSION': '9.9'}):
       self.check_working([EMCC])
 
@@ -685,12 +706,18 @@ fi
     # with no binaryen root, an error is shown
     restore_and_set_up()
 
+    # Remove wasm-opt from PATH so config setup cannot accidentally succeed by
+    # virtue of locating it in PATH.
+    env = os.environ.copy()
+    env['PATH'] = path_without_tool(env['PATH'], 'wasm-opt')
+
     open(EM_CONFIG, 'a').write('\nBINARYEN_ROOT = ""\n')
-    self.check_working([EMCC, test_file('hello_world.c')], 'BINARYEN_ROOT is set to empty value in %s' % EM_CONFIG)
+    self.check_working([EMCC, test_file('hello_world.c')], 'BINARYEN_ROOT is set to empty value in %s' % EM_CONFIG, env=env)
 
     open(EM_CONFIG, 'a').write('\ndel BINARYEN_ROOT\n')
-    self.check_working([EMCC, test_file('hello_world.c')], 'BINARYEN_ROOT not set in config (%s), and `wasm-opt` not found in PATH' % EM_CONFIG)
+    self.check_working([EMCC, test_file('hello_world.c')], 'BINARYEN_ROOT not set in config (%s), and `wasm-opt` not found in PATH' % EM_CONFIG, env=env)
 
+  @no_windows('Test relies on Unix-specific make_fake_tool')
   def test_empty_config(self):
     restore_and_set_up()
     make_fake_tool(self.in_dir('fake', 'wasm-opt'), 'foo')
@@ -701,6 +728,7 @@ fi
     with env_modify({'PATH': self.in_dir('fake') + os.pathsep + os.environ['PATH']}):
       self.check_working([EMCC])
 
+  @no_windows('Test relies on Unix-specific make_fake_tool')
   def test_missing_config(self):
     restore_and_set_up()
     make_fake_tool(self.in_dir('fake', 'wasm-opt'), 'foo')
@@ -782,6 +810,13 @@ fi
     self.run_process([EMBUILDER, 'build', f'{external_port_path}:dependency=sdl2', '--force'])
     self.assertExists(os.path.join(config.CACHE, 'sysroot', 'lib', 'wasm32-emscripten', 'lib_external-sdl2.a'))
 
+  def test_embuilder_transitive_pic(self):
+    restore_and_set_up()
+    self.run_process([EMBUILDER, 'clear', 'sdl2*'])
+    self.run_process([EMBUILDER, '--pic', 'clear', 'sdl2*'])
+    self.run_process([EMCC, '-sMAIN_MODULE=2', '-sUSE_SDL=2', '-sUSE_SDL_GFX=2', test_file('hello_world.c')])
+
+  @no_windows('Test relies on Unix-specific make_fake_tool')
   def test_binaryen_version(self):
     restore_and_set_up()
     with open(EM_CONFIG, 'a') as f:
@@ -799,12 +834,11 @@ fi
 
     # Touching package.json should cause compiler to fail with bootstrap message
     Path(utils.path_from_root('package.json')).touch()
-    err = self.expect_fail([EMCC, test_file('hello_world.c')])
-    self.assertContained('emcc: error: emscripten setup is not complete ("npm packages" is out-of-date). Run `bootstrap` to update', err)
+    expected = 'emcc: error: emscripten setup is not complete ("npm packages" is out-of-date). Run `bootstrap` to update'
+    self.assert_fail([EMCC, test_file('hello_world.c')], expected)
 
     # Running bootstrap.py should fix that
-    bootstrap = shared.bat_suffix(shared.path_from_root('bootstrap'))
-    self.run_process([bootstrap])
+    self.run_process([utils.exe_path_from_root('bootstrap')])
 
     # Now the compiler should work again
     self.run_process([EMCC, test_file('hello_world.c')])
@@ -820,8 +854,7 @@ fi
 
     # Remove from PATH every directory that contains clang.exe so that bootstrap.py cannot
     # accidentally succeed by virtue of locating tools in PATH.
-    new_path = [d for d in env['PATH'].split(os.pathsep) if not os.path.isfile(os.path.join(d, utils.exe_suffix('clang')))]
-    env['PATH'] = os.pathsep.join(new_path)
+    env['PATH'] = path_without_tool(env['PATH'], 'clang')
 
     # Running bootstrap.py should not fail
-    self.run_process([shared.bat_suffix(shared.path_from_root('bootstrap'))], env=env)
+    self.run_process([utils.exe_path_from_root('bootstrap')], env=env)
