@@ -3,11 +3,10 @@
 # University of Illinois/NCSA Open Source License.  Both these licenses can be
 # found in the LICENSE file.
 
-import logging
 import os
 
-TAG = 'version_1'
-HASH = '99bee75beb662f8520bbb18ad6dbf8590d30eb3a7360899f0ac4764ca72fe8013da37c9df21e525f9d2dc5632827d4b4cea558cbc938e7fbed0c41a29a7a2dc5'
+VERSION = '1.3.7'
+HASH = '10edd193f44f2b2d6085672e257b5be11066f5db27348d4cab53789bf9e8739c2df2cd7f42b83e4edbab53e10b5dc84320d6aab1deac0d8e18196a8a88b19b77'
 
 deps = ['ogg']
 
@@ -17,30 +16,26 @@ def needed(settings):
 
 
 def get(ports, settings, shared):
-  ports.fetch_project('vorbis', 'https://github.com/emscripten-ports/vorbis/archive/' + TAG + '.zip', 'Vorbis-' + TAG, sha512hash=HASH)
+  ports.fetch_project('vorbis', f'https://github.com/xiph/vorbis/releases/download/v{VERSION}/libvorbis-{VERSION}.zip', sha512hash=HASH)
 
   def create(final):
-    logging.info('building port: vorbis')
-    source_path = os.path.join(ports.get_dir(), 'vorbis', 'Vorbis-' + TAG)
-    ports.install_header_dir(os.path.join(source_path, 'include', 'vorbis'))
+    source_path = ports.get_dir('vorbis', 'libvorbis-' + VERSION)
+    ports.install_headers(os.path.join(source_path, 'include', 'vorbis'), target='vorbis')
+    ports.make_pkg_config('vorbis', VERSION, '-sUSE_VORBIS')
     ports.build_port(os.path.join(source_path, 'lib'), final, 'vorbis',
                      flags=['-sUSE_OGG'],
                      exclude_files=['psytune', 'barkmel', 'tone', 'misc'])
 
-  return [shared.Cache.get_lib('libvorbis.a', create)]
+  return [shared.cache.get_lib('libvorbis.a', create)]
 
 
 def clear(ports, settings, shared):
-  shared.Cache.erase_lib('libvorbis.a')
+  shared.cache.erase_lib('libvorbis.a')
 
 
 def process_dependencies(settings):
   settings.USE_OGG = 1
 
 
-def process_args(ports):
-  return []
-
-
 def show():
-  return 'vorbis (-sUSE_VORBIS; zlib license)'
+  return 'vorbis (-sUSE_VORBIS or --use-port=vorbis; zlib license)'
